@@ -1,16 +1,13 @@
-from operator import index
 import torch
 import copy
 import numpy as np
 import torch.fx
 import inspect
 import scipy, scipy.sparse
-import logging
 import copy
 import types
-
-
-_log = logging.getLogger(__name__)
+import random
+import warnings
 
 
 # ++++++++++++++++++++++ Core implementation ++++++++++++++++++++++
@@ -34,7 +31,7 @@ def set_dispatch_failure(action):
 # ====================== Dispatch defaults ======================
 
 
-class DispatchError(Exception):
+class DispatchError(Warning):
     pass
 
 
@@ -68,7 +65,7 @@ class SparseTensorWrapper(torch.Tensor):
         if grad_fmt is None:
             grad_fmt = (KeepAll(), torch.Tensor, KeepAll(), DenseTensor)
         return SparseTensorWrapper(
-            wrapped_tensor=wrapped,
+            wrapped_tensor_container=[wrapped],
             requires_grad=dense.requires_grad,
             grad_fmt=grad_fmt,
             dtype=dense.dtype,
@@ -77,19 +74,30 @@ class SparseTensorWrapper(torch.Tensor):
 
     def __new__(
         cls,
-        wrapped_tensor,
+        wrapped_tensor_container,
         requires_grad,
         grad_fmt,
         dtype,
         device,
+        *,
+        dummy_shape=None,
     ):
-        assert not isinstance(wrapped_tensor, SparseTensorWrapper)
+        assert not isinstance(wrapped_tensor_container[0], SparseTensorWrapper)
+
+        # We randomly initialize tensor to help finding bugs with the incorrect use of tensor data.
+        # We try to keep different shapes, but the same number of elements to avoid memory bugs that
+        # may be even harder to catch.
+        if dummy_shape is None:
+            dummy_shape = [1] * random.randint(5, 10) + [2, 2, 3]
+            random.shuffle(dummy_shape)
+        dummy = torch.randint(7770, 7779, dummy_shape, dtype=dtype, device=device)
+
         tensor = torch.Tensor._make_subclass(
             cls,
-            torch.tensor([987654321], dtype=dtype, device=device),
+            dummy,
             requires_grad,
         )
-        tensor.wrapped_tensor = wrapped_tensor
+        tensor._wrapped_tensor_container = wrapped_tensor_container
         tensor.update_grad_fix_hook()
         tensor.grad_fmt = grad_fmt
         return tensor
@@ -98,7 +106,7 @@ class SparseTensorWrapper(torch.Tensor):
         # shallow copy
         return sparse_tensor_builder(
             type(self),
-            self.wrapped_tensor,
+            self._wrapped_tensor_container,
             self.requires_grad,
             self.grad_fmt,
             self.dtype,
@@ -111,12 +119,24 @@ class SparseTensorWrapper(torch.Tensor):
         else:
             result = sparse_tensor_builder(
                 type(self),
-                copy.deepcopy(self.wrapped_tensor),
+                copy.deepcopy(self._wrapped_tensor_container),
                 copy.deepcopy(self.requires_grad),
                 copy.deepcopy(self.grad_fmt),
                 copy.deepcopy(self.dtype),
                 copy.deepcopy(self.device),
             )
+            if self.grad is not None:
+                assert isinstance(self.grad, SparseTensorWrapper)
+                grad_copy = copy.deepcopy(self.grad)
+                result.grad = SparseTensorWrapper(
+                    grad_copy.wrapped_tensor_container,
+                    grad_copy.requires_grad,
+                    grad_copy.grad_fmt,
+                    grad_copy.dtype,
+                    grad_copy.device,
+                    dummy_shape=get_dummy_shape(result),
+                )
+
             memo[id(self)] = result
             return result
 
@@ -125,7 +145,7 @@ class SparseTensorWrapper(torch.Tensor):
         if isinstance(self, SparseTensorWrapper):
             args = (
                 type(self),
-                self.wrapped_tensor,
+                self._wrapped_tensor_container,
                 self.requires_grad,
                 self.grad_fmt,
                 self.dtype,
@@ -161,10 +181,18 @@ class SparseTensorWrapper(torch.Tensor):
         except Exception as e:
             # Keep this error message to have an opportunity to catch the error
             # if exception is not reraised.
-            _log.error(f"Exception raised during handling __torch_function__: {e}")
+            warnings.warn(f"Exception raised during handling __torch_function__: {e}")
             # Sometimes this exception is silently ignored by PyTorch.
             raise e
 
+    @property
+    def wrapped_tensor(self):
+        return self._wrapped_tensor_container[0]
+
+    @wrapped_tensor.setter
+    def wrapped_tensor(self, value):
+        self._wrapped_tensor_container[0] = value
+
 
 class SparseParameterWrapper(SparseTensorWrapper, torch.nn.parameter.Parameter):
     @staticmethod
@@ -180,7 +208,7 @@ class SparseParameterWrapper(SparseTensorWrapper, torch.nn.parameter.Parameter):
         assert isinstance(sparse_tensor_wrapper, SparseTensorWrapper)
         return super().__new__(
             cls,
-            wrapped_tensor=sparse_tensor_wrapper.wrapped_tensor,
+            wrapped_tensor_container=sparse_tensor_wrapper._wrapped_tensor_container,
             requires_grad=sparse_tensor_wrapper.requires_grad,
             grad_fmt=sparse_tensor_wrapper.grad_fmt,
             dtype=sparse_tensor_wrapper.dtype,
@@ -212,15 +240,49 @@ class SparseParameterWrapper(SparseTensorWrapper, torch.nn.parameter.Parameter):
     torch.Tensor.grad.__get__,
     torch.Tensor.is_leaf.__get__,
     torch.Tensor.grad_fn.__get__,
-    torch.Tensor.backward,
     torch.Tensor.dtype.__get__,
     torch.Tensor.device.__get__,
     torch.Tensor.__hash__,  # returns id(self) by default
+    torch.Tensor.is_sparse.__get__,  # pretend to be dense
 )
 def sparse_default_impl(base_impl, func, types, *args, **kwargs):
     return base_impl(func, types, args, kwargs)
 
 
+def get_dummy_shape(tensor):
+    assert isinstance(tensor, SparseTensorWrapper)
+    # ugly hack to use base class dummy implementation
+    old_class = tensor.__class__
+    tensor.__class__ = torch.Tensor
+    dummy_shape = tensor.shape
+    tensor.__class__ = old_class
+    return dummy_shape
+
+
+@implements(
+    torch.Tensor.backward,
+)
+def sparse_backward_impl(base_impl, func, types, *args, **kwargs):
+    # here we try to match dummy shape of gradient to the shape of self before continuing
+    self = args[0]
+    grad = args[1] if len(args) > 1 else kwargs["gradient"]
+    match_grad = SparseTensorWrapper(
+        grad._wrapped_tensor_container,
+        grad.requires_grad,
+        grad.grad_fmt,
+        grad.dtype,
+        grad.device,
+        dummy_shape=get_dummy_shape(self),
+    )
+    new_args = [a for a in args]
+    if len(new_args) > 1:
+        new_args[1] = match_grad
+    new_kwargs = {k: v for k, v in kwargs.items()}
+    if "gradient" in new_kwargs:
+        new_kwargs["gradient"] = match_grad
+    return base_impl(func, types, new_args, new_kwargs)
+
+
 def sparse_redirect_impl(base_impl, func, _types, *args, **kwargs):
     (
         same_tensor,
@@ -231,11 +293,7 @@ def sparse_redirect_impl(base_impl, func, _types, *args, **kwargs):
     ) = check_op_semantics(func, args, kwargs)
 
     if num_inputs == 1:
-        [self] = [
-            x
-            for x in list(args) + list(kwargs.values())
-            if isinstance(x, SparseTensorWrapper)
-        ]
+        self, *other_args = args
         if func.__name__ == "__get__":
             # attribute access, try to redirect to wrapped tensor (e.g. ten.shape)
             attribute_name = func.__self__.__name__
@@ -245,11 +303,14 @@ def sparse_redirect_impl(base_impl, func, _types, *args, **kwargs):
             # call method of a class (e.g. ten.size())
             method_name = func.__name__
             wrapper_class = type(self.wrapped_tensor)
-            if hasattr(wrapper_class, method_name):
-                return getattr(wrapper_class, method_name)(*args, **kwargs)
 
-    _log.warning(
-        f"Using fallback dense implementation for read-only access without tensor output: {torch.overrides.resolve_name(func)}"
+            if hasattr(wrapper_class, method_name):
+                return getattr(wrapper_class, method_name)(
+                    self.wrapped_tensor, *other_args, **kwargs
+                )
+
+    warnings.warn(
+        f"Using fallback dense implementation for read-only access without tensor output: {torch.overrides.resolve_name(func)}", DispatchError
     )
     d_args = densify(args)
     d_kwargs = densify(kwargs)
@@ -264,10 +325,12 @@ def make_new_tensor_with_data_sharing(base_impl, func, types, *args, **kwargs):
 
 
 def sparse_tensor_builder(
-    wrapper_type, wrapped_tensor, requires_grad, grad_fmt, dtype, device
+    wrapper_type, wrapped_tensor_container, requires_grad, grad_fmt, dtype, device
 ):
     assert issubclass(wrapper_type, SparseTensorWrapper)
-    result = SparseTensorWrapper(wrapped_tensor, requires_grad, grad_fmt, dtype, device)
+    result = SparseTensorWrapper(
+        wrapped_tensor_container, requires_grad, grad_fmt, dtype, device
+    )
     if wrapper_type == SparseParameterWrapper:
         result = SparseParameterWrapper(result)
     return result
@@ -350,8 +413,9 @@ def check_op_semantics(op, args, kwargs):
     if op in OP_SEMANTICS_CACHE:
         return OP_SEMANTICS_CACHE[op]
 
-    _log.warning(
-        f"Semantics of {torch.overrides.resolve_name(op)} is unknown, trying to discover it by executing..."
+    warnings.warn(
+        f"Semantics of {torch.overrides.resolve_name(op)} is unknown, trying to discover it by executing...",
+        DispatchError
     )
 
     dargs = rand_as_args(args)
@@ -405,8 +469,9 @@ def sparse_fallback(base_impl, func, types, *args, **kwargs):
     ) = check_op_semantics(func, args, kwargs)
 
     if (same_tensor == {0: 0}) and (inplace_changes == [0]) and (num_outputs == 1):
-        _log.warning(
-            f"Using fallback dense implementation for inplace operation: {torch.overrides.resolve_name(func)}"
+        warnings.warn(
+            f"Using fallback dense implementation for inplace operation: {torch.overrides.resolve_name(func)}",
+            DispatchError,
         )
         # inplace operator that returns self (e.g. torch.Tensor.add_, torch.Tensor.copy_)
         d_args = densify(args)
@@ -548,25 +613,28 @@ def get_direct_name(fully_qualified_tensor_name):
     return fully_qualified_tensor_name.rsplit(".", 1)[-1]
 
 
+class TracedSubmodules:
+    def __init__(self):
+        self.data = {}
+
+    def access(self, module, submodule_path):
+        submodule = module
+        tokens = [x for x in submodule_path.split(".") if x]
+        for token in tokens:
+            submodule = getattr(submodule, token)
+        if submodule_path not in self.data:
+            self.data[submodule_path] = torch.fx.symbolic_trace(submodule)
+        return self.data[submodule_path]
+
+
 class SparsityBuilder:
-    def __init__(self, module):
-        self.module = module
-        self.traced_submodules = {}
+    def __init__(self):
         self.initial_weight_sparsifiers = {}
         self.weights = {}
         self.grad_weights = {}
         self.interms = {}
         self.grad_interms = {}
 
-    def traced_submodule(self, submodule_path):
-        submodule = self.module
-        tokens = [x for x in submodule_path.split(".") if x]
-        for token in tokens:
-            submodule = getattr(submodule, token)
-        if submodule_path not in self.traced_submodules:
-            self.traced_submodules[submodule_path] = torch.fx.symbolic_trace(submodule)
-        return self.traced_submodules[submodule_path]
-
     def set_weight(
         self,
         name,
@@ -657,8 +725,8 @@ class SparsityBuilder:
                 interm_name, KeepAll(), torch.Tensor, KeepAll(), DenseTensor
             )
 
-    def replace_with_traced_submodules(self, sparse_module):
-        for traced_module in self.traced_submodules.values():
+    def replace_with_traced_submodules(self, sparse_module, traced_submodules):
+        for traced_module in traced_submodules.values():
             traced_module.recompile()
 
         # start replacing submodules from the outermost to innermost
@@ -666,9 +734,7 @@ class SparsityBuilder:
             path, module = x
             return len([t for t in path.split(".") if t])
 
-        traced_submodules_list = sorted(
-            list(self.traced_submodules.items()), key=num_tokens
-        )
+        traced_submodules_list = sorted(list(traced_submodules.items()), key=num_tokens)
 
         for module_path, traced_module in traced_submodules_list:
             if module_path == "":
@@ -683,17 +749,20 @@ class SparsityBuilder:
 
         return sparse_module
 
-    def get_sparse_model(self):
+    def sparsify_model_inplace(self, module):
         self.fill_remaining()
+
+        traced_submodules = TracedSubmodules()
+
         with torch.no_grad():
-            sparse_module = copy.deepcopy(self.module)
+            sparse_module = module
 
             for name, (sp1, fmt1, sp2, fmt2) in self.interms.items():
                 grad_sp1, grad_fmt1, grad_sp2, grad_fmt2 = self.grad_interms[name]
 
                 direct_name = get_direct_name(name)
                 submodule_path = get_module_path(name)
-                submodule = self.traced_submodule(submodule_path)
+                submodule = traced_submodules.access(module, submodule_path)
 
                 [node] = (n for n in submodule.graph.nodes if n.name == direct_name)
                 node.target = sparsified_op(
@@ -702,7 +771,9 @@ class SparsityBuilder:
                     [(grad_sp1, grad_fmt1, grad_sp2, grad_fmt2)],
                 )
 
-            sparse_module = self.replace_with_traced_submodules(sparse_module)
+            sparse_module = self.replace_with_traced_submodules(
+                sparse_module, traced_submodules.data
+            )
 
             # weight modification should happen only after the call to replace_with_traced_submodules
             # otherwise these changes will be lost
@@ -725,6 +796,10 @@ class SparsityBuilder:
 
             return sparse_module
 
+    def get_sparse_model(self, module):
+        copied_module = copy.deepcopy(module)
+        return self.sparsify_model_inplace(copied_module)
+
 
 def get_format(tensor):
     if hasattr(tensor, "wrapped_tensor"):
@@ -783,8 +858,9 @@ def get_sparsifier_implementation(sparsifier, inp, out):
     if impl is None:
         err_msg = f"Sparsifier implementation is not registered. sparsifier: {sparsifier} inp: {inp} out: {out}."
         if DISPATCH_FAILURE == DISPATCH_WARN and inp != torch.Tensor:
-            _log.warning(
-                f"{err_msg} Use fallback implementation. {sparsifier} inp: {torch.Tensor} out: {out}"
+            warnings.warn(
+                f"{err_msg} Use fallback implementation. {sparsifier} inp: {torch.Tensor} out: {out}",
+                DispatchError,
             )
             # we can try to use fallback implementation input_format1 -> torch.Tensor -> sparsifier -> input_format2
             # instead of input_format1 -> sparsifier -> input_format2
@@ -884,6 +960,7 @@ PATCHED_OVERRIDES = {
     torch.stack: lambda tensors, dim=0, *, out=None: -1,
     torch.eq: lambda input, other, *, out=None: -1,
     torch.Tensor.eq: lambda input, other: -1,
+    torch.zeros_like: lambda input, *, dtype=None, layout=torch.strided, device=None, requires_grad=False, memory_format=None: -1,
 }
 
 
@@ -1003,114 +1080,136 @@ def create_failing_bwd_impl(exception):
     return failing_bwd_impl
 
 
+def find_gradient_fmt(tensor):
+    if isinstance(tensor, SparseTensorWrapper):
+        assert len(tensor.grad_fmt) == 4
+        return tensor.grad_fmt
+    if isinstance(tensor, torch.Tensor):
+        return (KeepAll(), torch.Tensor, KeepAll(), torch.Tensor)
+    return (None, None, None, None)
+
+
+def get_dummy_shapes(tensors):
+    dummy_shapes = []
+    for x in tensors:
+        if isinstance(x, SparseTensorWrapper):
+            dummy_shapes.append(get_dummy_shape(x))
+        else:
+            dummy_shapes.append(None)
+    return dummy_shapes
+
+
+def sparsifier_obj_to_classes(sparsifiers, formats):
+    return tuple(
+        ((s.__class__, f) if s is not None else None)
+        for s, f in zip(sparsifiers, formats)
+    )
+
+
 class SparseOperatorDispatcher(torch.autograd.Function):
     @staticmethod
     def forward(ctx, disp_state, *args_kwargs):
-        def find_gradient_fmt(tensor):
-            if isinstance(tensor, SparseTensorWrapper):
-                assert len(tensor.grad_fmt) == 4
-                return tensor.grad_fmt
-            if isinstance(tensor, torch.Tensor):
-                return (KeepAll(), torch.Tensor, KeepAll(), torch.Tensor)
-            return (None, None, None, None)
-
-        inp_fmt = tuple(get_format(a) for a in args_kwargs)
+        ctx.dummy_input_shapes = get_dummy_shapes(args_kwargs)
+        ctx.inp_fmt = tuple(get_format(a) for a in args_kwargs)
         ctx.grad_inp_fmt = tuple(find_gradient_fmt(a) for a in args_kwargs)
         ctx.disp_state = disp_state
-        sp1, fmt1, sp2, fmt2 = tuple(zip(*disp_state.out_fmt))
-        gsp1, gfmt1, gsp2, gfmt2 = tuple(zip(*disp_state.grad_out_fmt))
-        grad_inp_sp1, grad_inp_fmt1, grad_inp_sp2, grad_inp_fmt2 = tuple(
-            zip(*ctx.grad_inp_fmt)
-        )
-        op_out_fmt = tuple(
-            ((s.__class__, f) if s is not None else None) for s, f in zip(sp1, fmt1)
-        )
-        op_gout_fmt = tuple(
-            ((s.__class__, f) if s is not None else None) for s, f in zip(gsp2, gfmt2)
-        )
-        op_ginp_fmt = tuple(
-            ((s.__class__, f) if s is not None else None)
-            for s, f in zip(grad_inp_sp1, grad_inp_fmt1)
-        )
+        out_sp1, out_fmt1, out_sp2, out_fmt2 = zip(*ctx.disp_state.out_fmt)
+        op_out_fmt = sparsifier_obj_to_classes(out_sp1, out_fmt1)
 
         # warning: do not merge following section with try-except block for backward pass,
         # because their trivially_dense property may be different
-
-        fwd_impl_fallback = False
-
+        ctx.fwd_impl_fallback = False
         # find implementation for the forward pass
         try:
-            op_impl_fwd = get_fwd_op_impl(disp_state.orig_op, inp_fmt, op_out_fmt)
+            op_impl_fwd = get_fwd_op_impl(disp_state.orig_op, ctx.inp_fmt, op_out_fmt)
         except DispatchError as e:
             # are all types are trivially reducible to dense?
-            trivially_dense = check_dense_inputs(inp_fmt) and check_dense_outputs(
+            trivially_dense = check_dense_inputs(ctx.inp_fmt) and check_dense_outputs(
                 op_out_fmt
             )
 
             if not trivially_dense:
-                _log.warning(f"{e}\nFallback to dense implementation.")
+                warnings.warn(f"{e}\nFallback to dense implementation.", DispatchError)
                 if DISPATCH_FAILURE == DISPATCH_RAISE:
                     raise e
 
             # try to create fallback implementation
+            ctx.fwd_impl_fallback = True
+            op_impl_fwd = create_fallback_fwd_impl(out_fmt1)
 
-            fwd_impl_fallback = True
-            op_impl_fwd = create_fallback_fwd_impl(fmt1)
-
-        # find implementation for the backward pass
-        try:
-            ctx.op_impl_bwd = get_bwd_op_impl(
-                disp_state.orig_op, gfmt2, op_ginp_fmt, inp_fmt
-            )
-        except DispatchError as e:
-            # are all types are trivially reducible to dense?
-            trivially_dense = (
-                check_dense_inputs(gfmt2)
-                and check_dense_outputs(op_ginp_fmt)
-                and check_dense_inputs(inp_fmt)
-            )
-
-            if trivially_dense:
-                ctx.op_impl_bwd = create_fallback_bwd_impl(grad_inp_fmt1)
-            else:
-                if DISPATCH_FAILURE == DISPATCH_RAISE:
-                    raise e
-                else:
-                    if fwd_impl_fallback:
-                        _log.warning(f"{e}\nFallback to dense implementation.")
-                        ctx.op_impl_bwd = create_fallback_bwd_impl(grad_inp_fmt1)
-                    else:
-                        _log.warning(
-                            f"{e}\nCan't fallback to dense backward implementation because custom forward implementation was used. Attempt to compute backward will raise an exception."
-                        )
-                        ctx.op_impl_bwd = create_failing_bwd_impl(e)
-
-        tmp_outputs = op_impl_fwd(ctx, args_kwargs, sp1)
+        tmp_outputs = op_impl_fwd(ctx, args_kwargs, out_sp1)
         tmp_outputs = canonicalize_tensor_tuple(tmp_outputs)
-        check_formats(f"{disp_state.orig_op} (fwd)", tmp_outputs, fmt1)
+        check_formats(f"{disp_state.orig_op} (fwd)", tmp_outputs, out_fmt1)
         assert len(tmp_outputs) == len(disp_state.out_fmt)
         outputs = tuple(
             get_sparsifier_implementation(s2.__class__, f1, f2)(s2, tmp_out)
-            for tmp_out, f1, s2, f2 in zip(tmp_outputs, fmt1, sp2, fmt2)
+            for tmp_out, f1, s2, f2 in zip(tmp_outputs, out_fmt1, out_sp2, out_fmt2)
         )
         outptus = simplify_tensor_tuple(outputs)
         return outptus
 
     @staticmethod
     def backward(ctx, *args):
-        sp1, fmt1, sp2, fmt2 = tuple(zip(*ctx.grad_inp_fmt))
-        grad_out_fmt = tuple(get_format(a) for a in args)
-        op_inp_fmt = tuple(
-            ((s.__class__, f) if s is not None else None) for s, f in zip(sp1, fmt1)
-        )
-        tmp_grad_inputs = ctx.op_impl_bwd(ctx, args, sp1)
+        grad_out_fmt_actual = tuple(get_format(a) for a in args)
+        gout_sp1, gout_fmt1, gout_sp2, gout_fmt2 = zip(*ctx.disp_state.grad_out_fmt)
+
+        if gout_fmt2 != grad_out_fmt_actual:
+            raise ValueError(
+                f"Backward implementation received output gradients of incorrect format. Expected: {ctx.disp_state.grad_out_fmt}. Actual: {grad_out_fmt_actual}."
+            )
+
+        ginp_sp1, ginp_fmt1, ginp_sp2, ginp_fmt2 = zip(*ctx.grad_inp_fmt)
+
+        op_ginp_fmt = sparsifier_obj_to_classes(ginp_sp1, ginp_fmt1)
+
+        # find implementation for the backward pass
+        try:
+            op_impl_bwd = get_bwd_op_impl(
+                ctx.disp_state.orig_op, gout_fmt2, op_ginp_fmt, ctx.inp_fmt
+            )
+        except DispatchError as e:
+            # are all types are trivially reducible to dense?
+            trivially_dense = (
+                check_dense_inputs(gout_fmt2)
+                and check_dense_outputs(op_ginp_fmt)
+                and check_dense_inputs(ctx.inp_fmt)
+            )
+
+            if trivially_dense:
+                op_impl_bwd = create_fallback_bwd_impl(ginp_fmt1)
+            else:
+                if DISPATCH_FAILURE == DISPATCH_RAISE:
+                    raise e
+                else:
+                    if ctx.fwd_impl_fallback:
+                        op_impl_bwd = create_fallback_bwd_impl(ginp_fmt1)
+                    else:
+                        op_impl_bwd = create_failing_bwd_impl(e)
+
+        tmp_grad_inputs = op_impl_bwd(ctx, args, ginp_sp1)
         tmp_grad_inputs = canonicalize_tensor_tuple(tmp_grad_inputs)
-        check_formats(f"{ctx.disp_state.orig_op} (bwd)", tmp_grad_inputs, fmt1)
+        check_formats(f"{ctx.disp_state.orig_op} (bwd)", tmp_grad_inputs, ginp_fmt1)
         grad_inputs = tuple(
             get_sparsifier_implementation(s2.__class__, f1, f2)(s2, inp)
-            for inp, f1, s2, f2 in zip(tmp_grad_inputs, fmt1, sp2, fmt2)
+            for inp, f1, s2, f2 in zip(tmp_grad_inputs, ginp_fmt1, ginp_sp2, ginp_fmt2)
         )
-        return (None, *grad_inputs)
+        # make shapes of grad_inputs the same as shapes of inputs to avoid complaints from autograd
+        reshaped_grad_inputs = []
+        for x, ds in zip(grad_inputs, ctx.dummy_input_shapes):
+            if isinstance(x, SparseTensorWrapper):
+                rx = SparseTensorWrapper(
+                    wrapped_tensor_container=x._wrapped_tensor_container,
+                    requires_grad=x.requires_grad,
+                    grad_fmt=x.grad_fmt,
+                    dtype=x.dtype,
+                    device=x.device,
+                    dummy_shape=ds,
+                )
+                reshaped_grad_inputs.append(rx)
+            else:
+                reshaped_grad_inputs.append(x)
+
+        return (None, *reshaped_grad_inputs)
 
 
 class TensorIdx:
